/****************************************************************************

    spectromatic

    Copyright (C) 1997-2002 Daniel Franklin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

 ****************************************************************************/

#include <stdio.h>
#include <malloc.h>
#include <png.h>
#include <string.h>
#include <wave.h>
#include <math.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_fft_real.h>
#include <gsl/gsl_fft_halfcomplex.h>
#define _GNU_SOURCE
#include <getopt.h>

#define VERSION "1.0"

void usage (char *pname)
{
	fprintf (stderr,
"
Usage: %s [options] file.wav

Options:
	--window <size>		FFT window size (default: 2048)
	
	--step <size>		Slide the window by <size> samples each time 
				(default: 200)
	
	--inverse		White/Colour-on-black
	
	--logarithmic		Intensity is calculated by
				log10 (1 + 9 * (amplitude / max_amp)) -
				better for visualising weaker signals

	--combine XYZ		Determines how a stereo signal is
				represented as an RGB image. The argument is
				a 3-character block, each character
				corresponding to R, G and B components. It
				may take values of L, R, A, D, X or Z,
				meaning Left, Right, Average, Difference,
				Cross-Correlation or Zero respectively.

	--verbose		Slightly less quiet
	
	--version		Print out the version
	
	--help			You're in it.

", pname);
	

	exit (-1);
}

int main (int argc, char **argv)
{
	char *data, *p;
	char R = 'L', G = 'D', B = 'R';
	double *ch1 = NULL, *ch2 = NULL, *mags = NULL;
	waveheaderstruct header;
	FILE *f;
	int i, j;
	int window_size = 2048, step_size = 200;
	double ch1max = 0, ch2max = 0;
	int rows = 0;
	png_structp png_ptr;
	png_infop info_ptr;
	png_bytep array;
	png_byte left, right;
	int rowbytes;
	int inverse = 0, verbose = 0, logarithmic = 0;
	char outfname [256];
	FILE *buffer, *png;
	int c;
	
	while (1) {
		int option_index = 0;
		
		static struct option long_options[] = {
			{"window", 1, 0, 'w'},
			{"step", 1, 0, 's'},
			{"inverse", 0, 0, 'i'},
			{"logarithmic", 0, 0, 'l'},
			{"combine", 1, 0, 'c'},
			{"verbose", 0, 0, 'v'},
			{"version", 0, 0, 'V'},
			{"help", 0, 0, 'h'},
			{0, 0, 0, 0}
		};
		
		c = getopt_long (argc, argv, "w:s:ilc:vhV", long_options, &option_index);

		if (c == -1)
			break;
			
		switch (c) {
		case 0:
			printf ("option %s", long_options [option_index].name);
			
			if (optarg)
				printf (" with arg %s", optarg);
			
			printf ("\n");
			break;

		case 'w':
			if (sscanf (optarg, "%i", &window_size) != 1 || window_size <= 0) {
				fprintf (stderr, "Error: window size must be a positive integer\n");
				exit (-1);
			}
			
			break;
			
		case 's':
			if (sscanf (optarg, "%i", &step_size) != 1 || step_size <= 0) {
				fprintf (stderr, "Error: step size must be a positive integer\n");
				exit (-1);
			}
			
			break;
			
		case 'i':
			inverse = 1;
			break;
			
		case 'c':
			if (strlen (optarg) != 3 || 
					!strchr ("LRADXZ", optarg [0]) || 
					!strchr ("LRADXZ", optarg [1]) || 
					!strchr ("LRADXZ", optarg [2])) {
				fprintf (stderr, "Error: - argument to --combine should contain 3 characters in the set LRSDXZ\n");
				exit (-1);
			}
			
			R = optarg [0];
			G = optarg [1];
			B = optarg [2];
			
			break;

		case 'v':
			verbose = 1;
			break;
			
		case 'l':
			logarithmic = 1;
			break;
			
		case 'V':
			fprintf (stderr, "Spectromatic version %s\n", VERSION);
			exit (0);
			break;
			
		case 'h':
		case '?':
		default:
			usage (argv [0]);
		}
	}
	
	if (optind >= argc)
		usage (argv [0]);
	
	while (optind < argc) {
		f = fopen (argv [optind], "r");
	
		if (!f) {
			fprintf (stderr, "Error - could not open file %s for reading.\n", argv [optind]);
			exit (-1);
		}
		
		read_header (f, &header);
		
		if (verbose)
			print_header (f);
		
		if (header.modus != 1 && header.modus != 2) {
			fprintf (stderr, "Error: %s is neither mono nor stereo...\n", argv [optind]);
			exit (-1);
		}

		if (header.byte_p_spl / header.modus != 2) {
			fprintf (stderr, "Error: %s is not 16-bit...\n", argv [optind]);
			exit (-1);
		}

		if (!(buffer = tmpfile ())) {
			fprintf (stderr, "Error - cannot create tempfile\n");
			exit (-2);
		}
		
		data = (char*) malloc (window_size * header.byte_p_spl);
		mags = (double*) malloc (window_size / 2 * sizeof (double));
		ch1 = (double*) malloc (window_size * sizeof (double));

		if (header.modus == 2)
			ch2 = (double*) malloc (window_size * sizeof (double));

		fprintf (stderr, "samples = %li, step_size = %i, window_size = %i\n", header.data_length / header.byte_p_spl, step_size, window_size);

		for (i = -window_size; i < window_size + (int)(header.data_length / header.byte_p_spl); i += step_size) {
			read_frames (f, i, window_size, data);
			
			if (header.modus == 1) {
				for (j = 0; j < window_size; j++)
					ch1 [j] = data [j << 1] + (data [(j << 1) + 1] << 8);
	
			} else {
				for (j = 0; j < window_size; j++) {
					ch1 [j] = data [j << 2] + (data [(j << 2) + 1] << 8);
					ch2 [j] = data [(j << 2) + 2] + (data [(j << 2) + 3] << 8);
				}
				
				gsl_fft_real_radix2_transform (ch2, 1, window_size);
			}
			
			gsl_fft_real_radix2_transform (ch1, 1, window_size);
				
			mags [0] = fabs (ch1 [0]);
				
			for (j = 0; j < window_size / 2; j++) {
				mags [j] = sqrt (ch1 [j] * ch1 [j] + ch1 [window_size - j] * ch1 [window_size - j]);
			
				if (mags [j] > ch1max)
					ch1max = mags [j];
			}
			
			if (!fwrite ((void *) mags, sizeof (double), window_size / 2, buffer)) {
				fprintf (stderr, "Error: cannot write to temp file: %s\n", strerror (errno));
				exit (-1);
			}

			if (header.modus == 2) {
				mags [0] = fabs (ch2 [0]);
				
				for (j = 0; j < window_size / 2; j++) {
					mags [j] = sqrt (ch2 [j] * ch2 [j] + ch2 [window_size - j] * ch2 [window_size - j]);

					if (mags [j] > ch2max)
						ch2max = mags [j];
				}
					
				if (!fwrite ((void *) mags, sizeof (double), window_size / 2, buffer)) {
					fprintf (stderr, "Error: cannot write to temp file: %s\n", strerror (errno));
					exit (-1);
				}
			}
			
			rows++;
		}

		free (data);
		free (mags);

		fclose (f);
	
		if (ch1max == 0) {
			fprintf (stderr, "Error - channel 1 maximum value = 0!\n");
			exit (-2);
		}

		if (header.modus == 2 && ch2max == 0) {
			fprintf (stderr, "Error - channel 2 maximum value = 0!\n");
			exit (-2);
		}
	
		rewind (buffer);
		
		strcpy (outfname, argv [optind]);
		p = strrchr (outfname, '.');
		
		if (p)
			*p = 0;
		
		strcat (outfname, ".png");
	
		if (!(png = fopen (outfname, "w"))) {
			fprintf (stderr, "Error - cannot open %s for writing\n", outfname);
			exit (-3);
		}

		png_ptr = png_create_write_struct (PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
		info_ptr = png_create_info_struct (png_ptr);

		png_set_IHDR (png_ptr, info_ptr, window_size / 2, rows, 8, header.modus == 1 ? PNG_COLOR_TYPE_GRAY : PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
		rowbytes = png_get_rowbytes (png_ptr, info_ptr);
	
		array = (png_bytep) png_malloc (png_ptr, rowbytes);
		png_init_io (png_ptr, png);
		png_write_info (png_ptr, info_ptr);

		if (header.modus == 1)	
			for (i = 0; i < rows; i++) {
				if (logarithmic) {
					fread (ch1, sizeof (double), window_size / 2, buffer);
					
					for (j = 0; j < window_size / 2; j++)
						array [j] = (png_byte)(255 * log10 (1 + 9 * (ch1 [j] / ch1max)));
				} else {
					fread (ch1, sizeof (double), window_size / 2, buffer);
					
					for (j = 0; j < window_size / 2; j++) {
						array [j] = (png_byte)(255 * (ch1 [j] / ch1max));
					}
				}
			
				if (!inverse) {
					for (j = 0; j < window_size / 2; j++) {
						array [j] = 255 - array [j];
					}
				}

				png_write_row (png_ptr, array);
			}
		else if (header.modus == 2)	
			for (i = 0; i < rows; i++) {
				fread (ch1, sizeof (double), window_size / 2, buffer);
				fread (ch2, sizeof (double), window_size / 2, buffer);
			
				if (logarithmic)
					for (j = 0; j < window_size / 2; j++) {
						left = (png_byte)(255 * log10 (1 + 9 * (ch1 [j] / ch1max)));
						right = (png_byte)(255 * log10 (1 + 9 * (ch2 [j] / ch1max)));
						
						switch (R) {
						case 'L':
							array [j * 3] = left;
							break;
							
						case 'R':
							array [j * 3] = right;
							break;
							
						case 'A':
							array [j * 3] = left + right / 2;
							break;
						
						case 'D':
							array [j * 3] = abs (left - right);
							break;

						case 'X':
							array [j * 3] = left * right / 256;
							break;

						case 'Z':
							array [j * 3] = 0;
							break;
						}

						switch (G) {
						case 'L':
							array [j * 3 + 1] = left;
							break;
							
						case 'R':
							array [j * 3 + 1] = right;
							break;
							
						case 'A':
							array [j * 3 + 1] = left + right / 2;
							break;
						
						case 'D':
							array [j * 3 + 1] = abs (left - right);
							break;

						case 'X':
							array [j * 3 + 1] = left * right / 256;
							break;

						case 'Z':
							array [j * 3 + 1] = 0;
							break;
						}
						
						switch (B) {
						case 'L':
							array [j * 3 + 2] = left;
							break;
							
						case 'R':
							array [j * 3 + 2] = right;
							break;
							
						case 'A':
							array [j * 3 + 2] = left + right / 2;
							break;
						
						case 'D':
							array [j * 3 + 2] = abs (left - right);
							break;

						case 'X':
							array [j * 3 + 2] = left * right / 256;
							break;

						case 'Z':
							array [j * 3 + 2] = 0;
							break;
						}
					}
				else
					for (j = 0; j < window_size / 2; j++) {
						left = (png_byte)(255 * (ch1 [j] / ch1max));
						right = (png_byte)(255 * (ch2 [j] / ch2max));
						
						switch (R) {
						case 'L':
							array [j * 3] = left;
							break;
							
						case 'R':
							array [j * 3] = right;
							break;
							
						case 'A':
							array [j * 3] = left + right / 2;
							break;
						
						case 'D':
							array [j * 3] = abs (left - right);
							break;

						case 'X':
							array [j * 3] = left * right / 256;
							break;

						case 'Z':
							array [j * 3] = 0;
							break;
						}

						switch (G) {
						case 'L':
							array [j * 3 + 1] = left;
							break;
							
						case 'R':
							array [j * 3 + 1] = right;
							break;
							
						case 'A':
							array [j * 3 + 1] = left + right / 2;
							break;
						
						case 'D':
							array [j * 3 + 1] = abs (left - right);
							break;

						case 'X':
							array [j * 3 + 1] = left * right / 256;
							break;

						case 'Z':
							array [j * 3 + 1] = 0;
							break;
						}
						
						switch (B) {
						case 'L':
							array [j * 3 + 2] = left;
							break;
							
						case 'R':
							array [j * 3 + 2] = right;
							break;
							
						case 'A':
							array [j * 3 + 2] = left + right / 2;
							break;
						
						case 'D':
							array [j * 3 + 2] = abs (left - right);
							break;

						case 'X':
							array [j * 3 + 2] = left * right / 256;
							break;

						case 'Z':
							array [j * 3 + 2] = 0;
							break;
						}
					}
			
				if (!inverse) {
					for (j = 0; j < window_size / 2; j++) {
						array [j * 3] = 255 - array [j * 3];
						array [j * 3 + 1] = 255 - array [j * 3 + 1];
						array [j * 3 + 2] = 255 - array [j * 3 + 2];
					}
				}

				png_write_row (png_ptr, array);
			}
	
		png_write_end (png_ptr, info_ptr);
	
		png_destroy_write_struct (&png_ptr, &info_ptr);

		free (array);
		free (ch1);
		
		if (header.modus == 2)
			free (ch2);
		
		fclose (buffer);
		fclose (png);

		optind++;
	}

	return 0;
}
